package org.linaro.mmtest;

import android.app.Activity;
import android.os.Bundle;
import android.os.Build;
import android.view.Window;
import android.util.Log;
import android.media.AudioManager;
import android.media.MediaPlayer;
import android.media.MediaPlayer.OnCompletionListener;
import android.os.Process;
import android.view.SurfaceView;
import android.app.AlertDialog;
import android.content.res.AssetManager;
import android.content.Context;
import android.widget.EditText;

import java.io.RandomAccessFile;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.io.File;
import java.lang.Math;
import java.lang.Runtime;
import java.util.List;

public class MultimediaTest extends Activity implements android.media.MediaPlayer.OnCompletionListener,android.media.MediaPlayer.OnErrorListener,android.media.MediaPlayer.OnInfoListener {
	private EditText status;
	private SurfaceView surfaceView;
	private static final String TAG = "MultimediaTest";
	private int currentEncodingTest = -1;
	private int currentTest = -1;
	private boolean ranAnyTest = false;
	private MediaPlayer mp;
	private String failure;
	private long timestamp;
	private long cpuTimestamp;
	private int expectedDuration;
	private final Test[] tests = {
		// Audio Codecs
		new Test("test-AAC-Stereo-96k.m4a", "AAC Stereo 96k", "-vn -acodec aac -strict experimental -ab 160k -ar 48000 -ac 2", false),
		new Test("test-AAC-Mono-8k.m4a", "AAC Mono 8k", "-vn -acodec aac -strict experimental -ab 32k -ar 8000 -ac 1", false),
		// AMR disabled because of licensing issues
		// new Test("test-AMR-NB.3gp", "AMR-NB", "", false),
		new Test("test-flac.flac", "FLAC", "-vn -acodec flac -ac 2 -ar 44100", false, 12 /*android.os.Build.HONEYCOMB_MR1*/),
		new Test("test-MP3.mp3", "MP3", "-vn -acodec libmp3lame -ab 64k", false),
		new Test("test-Vorbis.ogg", "Ogg Vorbis", "-vn -acodec libvorbis -f ogg -ab 64k", false),
		// no ffmpegParameters specified for the wav files because ffmpeg can't create those
		// A way to create them could be e.g.
		// mplayer $IN -benchmark -vc null -vo null -ao pcm:fast:waveheader:file=test-PCM.wav
		// sox -S test-PCM.wav -b 8 test-PCM8.wav
		// sox -S test-PCM.wav -b 16 test-PCM16.wav
		// But mplayer and sox are not (yet) available on Android, so we limit this to pre-existing files
		new Test("test-PCM8.wav", "PCM 8-Bit", "", false),
		new Test("test-PCM16.wav", "PCM 16-Bit", "", false),

		// Video Codecs
		new Test("test-H.263.3gp", "H.263", "-vcodec h263 -s cif -r 25 -vb 768000 -an -f 3gp"),
		new Test("test-H.264.m4v", "H.264", "-vcodec libx264 -vb 1024000 -vpre libx264-baseline -an -f mp4"),
		new Test("test-MPEG4_SP.mp4", "MPEG-4 SP", "-vcodec mpeg4 -vb 768000 -an -f mp4"),
		new Test("test-VP8.webm", "VP8", "-vcodec libvpx -vb 768000 -an -f webm"),

		// Mixed A/V
		new Test("test-H.263-AAC.3gp", "3GP(H.263/AAC)", "-vcodec h263 -s cif -r 25 -vb 768000 -acodec libfaac -ac 1 -ar 8000 -ab 32000 -f 3gp"),
		new Test("test-H.264-AAC.mp4", "MP4(H.264/AAC)", "-vcodec libx264 -vb 1024000 -vpre libx264-baseline -acodec aac -strict experimental -ab 96000 -f mp4"),
		new Test("test-H.264-AAC-HD.mp4", "HD-MP4(H.264/AAC)", "-s hd1080 -vcodec libx264 -vb 2048000 -vpre libx264-baseline -acodec aac -strict experimental -ab 96000 -f mp4"),
		new Test("test-MPEG4-MP3.mp4", "MP4(MPEG4/MP3)", "-vcodec mpeg4 -vb 768000 -acodec libmp3lame -ab 64000 -f mp4"),
		new Test("test-VP8-Vorbis.webm", "WebM(VP8/Vorbis)", "-vcodec libvpx -vb 1024000 -vpre libvpx-360p -f webm -acodec libvorbis -ab 48000", true, 10 /*android.os.Build.GINGERBREAD_MR1*/)
	};

	class Test {
		public String filename;
		public String name;
		public String ffmpegParameters;
		public boolean hasVideo;
		public int minVersion;

		public Test(String pFilename, String pName, String pFfmpeg, boolean pHasVideo, int pMinVersion) {
			filename = pFilename;
			name = pName;
			ffmpegParameters = pFfmpeg;
			hasVideo = pHasVideo;
			minVersion = pMinVersion;
		}

		public Test(String pFilename, String pName, String pFfmpeg, boolean pHasVideo) {
			this(pFilename, pName, pFfmpeg, pHasVideo, 0);
		}

		public Test(String pFilename, String pName, String pFfmpeg) {
			this(pFilename, pName, pFfmpeg, true, 0);
		}

		public Test(String pFilename, String pName) {
			this(pFilename, pName, "", true, 0);
		}
	}

	public void status(String s) {
		Log.v(TAG, s);
		status.setText(s);
	}

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);
		status = (EditText) findViewById(R.id.status);
		surfaceView = (SurfaceView) findViewById(R.id.surface_view);

		status("Extracting sample video from assets");
		if(!copySampleToFilesystem()) {
			status("FAILED copying sample video from assets");
			AlertDialog dlg=new AlertDialog.Builder(this).setTitle("Error").setMessage("An error occurred while trying to extract the sample video from assets").create();
			dlg.show();
			finish();
		}

		status("Starting encoding tests");

		status("Starting playback tests");
		mp=new MediaPlayer();
		mp.setOnCompletionListener(this);
		mp.setOnErrorListener(this);
		mp.setOnInfoListener(this);
		mp.setAudioStreamType(android.media.AudioManager.STREAM_MUSIC);
		mp.setDisplay(surfaceView.getHolder());

		runNextTest();
	}

	public void runNextTest() {
		failure = "";
		if(++currentEncodingTest <= tests.length) {
			Test t=tests[currentEncodingTest];
			status("Running encoding test: " + t.name);
			if(t.ffmpegParameters.length() == 0) {
				status("skipped, no encoder available");
				runNextTest();
				return;
			}
			Runtime rt = Runtime.getRuntime();
			try {
				timestamp = System.currentTimeMillis();
				String[] codecArgs = t.ffmpegParameters.split(" ");
				String[] ffmpegArgs = new String[codecArgs.length+5];
				ffmpegArgs[0] = "/system/bin/ffmpeg";
				ffmpegArgs[1] = "-y";
				ffmpegArgs[2] = "-i";
				ffmpegArgs[3] = getFilesDir().toString() + "/test-H.264-AAC-HD.mp4";
				System.arraycopy(codecArgs, 0, ffmpegArgs, 4, codecArgs.length);
				ffmpegArgs[4+codecArgs.length] = getFilesDir().toString() + "/" + t.filename;
				java.lang.Process ffmpeg = rt.exec(ffmpegArgs);
				// FIXME parse ffmpeg output instead of discarding it
				ffmpeg.getErrorStream().close();
				ffmpeg.getOutputStream().close();
				ffmpeg.getInputStream().close();
				ffmpeg.waitFor();
				long elapsedTime = System.currentTimeMillis() - timestamp;
				File f = new File(getFilesDir().toString() + "/" + t.filename);
				if(f.exists())
					Log.i(TAG, "PASS:" + t.name + ":encode");
				else
					Log.i(TAG, "FAIL:" + t.name + ":encode");
				f = null;
				Log.i(TAG, "PERF:Enc" + t.name + ":" + Long.toString(elapsedTime) + " ms");
			} catch(Throwable e) {
				status("FAILED: " + e.getMessage());
			}
			runNextTest();
			return;
		}
		if(++currentTest > tests.length) {
			status("Tests completed");
			if(!ranAnyTest) {
				AlertDialog dlg=new AlertDialog.Builder(this).setTitle("Files not found").setMessage("Multimedia files to be tested could not be found on the SD card. Please put the files (generated by the creator script) into the /mmtest directory of the SD card and try again.").create();
				dlg.show();
			}
			finish();
		}
		Test t=tests[currentTest];
		if(t.minVersion > android.os.Build.VERSION.SDK_INT) {
			Log.i(TAG, "SKIP:" + t.name + ":Not expected to pass on this version of Android");
			runNextTest();
			return;
		}
		status("Running playback test " + t.name);
		mp.reset();
		try {
			mp.setDataSource("/sdcard/mmtest/" + t.filename);
			mp.prepare();
			ranAnyTest = true;
		} catch(java.io.IOException e) {
			Log.e(TAG, "ERR:" + tests[currentTest].name + ":test:IO Error - test file seems to be missing: /sdcard/mmtest/" + t.filename);
			runNextTest();
			return;
		}
		if((mp.getVideoHeight() == 0) ^ t.hasVideo)
			Log.i(TAG, "PASS:" + tests[currentTest].name + ":sanityCheck:Test and Player agree about whether or not the content has a video part");
		else
			Log.i(TAG, "FAIL:" + tests[currentTest].name + ":sanityCheck:Test and Player disagree about whether or not the content has a video part");
		timestamp = System.currentTimeMillis();
		cpuTimestamp = cpuTime();
		expectedDuration = mp.getDuration();
		if(expectedDuration <= 0)
			Log.i(TAG, "WARN:" + tests[currentTest].name + ":Couldn't determine duration");
		mp.start();
	}

	@Override
	public void onCompletion(MediaPlayer mp) {
		long elapsedTime = System.currentTimeMillis() - timestamp;
		long elapsedCpuTime = cpuTime() - cpuTimestamp;
		if(failure.length() == 0) {
			if(expectedDuration >= 0 && Math.abs(elapsedTime - expectedDuration) > 10000) {
				Log.i(TAG, "WARN:" + tests[currentTest].name + ":Actual playing time (" + Long.toString(elapsedTime) + "ms) differs significantly from expected time (" + Integer.toString(expectedDuration) + "ms)");
			}
			Log.i(TAG, "PASS:" + tests[currentTest].name + ":playback");
		} else
			Log.i(TAG, "FAIL:" + tests[currentTest].name + ":playback:" + failure);
		// FIXME we should find a reasonable threshold to make
		// an educated guess on whether or not a codec is hardware
		// accelerated.
		// Will require some testing on different hardware...
		Log.i(TAG, "PERF:" + tests[currentTest].name + ":" + Long.toString(elapsedCpuTime) + " jiffies used");
		runNextTest();
	}

	@Override
	public boolean onError(MediaPlayer mp, int what, int extra) {
		failure = "Got error code " + Integer.toString(what) + "/" + Integer.toString(extra);
		return false; // call onCompletion
	}

	@Override
	public boolean onInfo(MediaPlayer mp, int what, int extra) {
		Log.i(TAG, "NOTE:" + tests[currentTest].name + ":playback:MediaPlayer sent info " + Integer.toString(what) + "/" + Integer.toString(extra));
		return true;
	}

	public long cpuTime() {
		long usedTime = 0;
		try {
			RandomAccessFile reader = new RandomAccessFile("/proc/" + Integer.toString(Process.myPid()) + "/stat", "r");
			String[] toks=reader.readLine().split(" ");
			// 13: jiffies in user mode
			// 14: jiffies in kernel mode
			// 15: jiffies in user mode of child processes (maybe some codecs launch an external decoder...)
			// 16: jiffies in kernel mode of child processes
			usedTime = Long.parseLong(toks[13]) + Long.parseLong(toks[14]) + Long.parseLong(toks[15]) + Long.parseLong(toks[16]);
			reader.close();
		} catch(java.io.FileNotFoundException e) {
			Log.e(TAG, "ERR:/proc not mounted?");
		} catch(java.io.IOException e) {
			Log.e(TAG, "ERR:I/O error while reading CPU load");
		}
		return usedTime;
	}

	public boolean copySampleToFilesystem() {
		try {
			AssetManager a = getAssets();
			FileOutputStream out = openFileOutput("test-H.264-AAC-HD.mp4", Context.MODE_WORLD_READABLE);
			InputStream in = a.open("big_buck_bunny_1080p_H264_AAC_25fps_7200K_short.mp4");
			byte[] buf = new byte[8192];
			int len;
			while((len = in.read(buf)) > 0) {
				out.write(buf, 0, len);
			}
			in.close();
			out.close();
			return true;
		} catch(java.io.FileNotFoundException e) {
			Log.e(TAG, "ERR:Can't read sample video from assets:" + e.getMessage());
		} catch(java.io.IOException e) {
			Log.e(TAG, "ERR:Can't copy sample video from assets:" + e.getMessage());
		}
		return false;
	}
}
